<?php
/*
 * API Export controller
 *
 * NOTICE OF LICENSE
 * This source file is subject to the Open Software License (OSL 3.0)
 * that is bundled with this package in the file LICENSE.TXT.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/OSL-3.0
 *
 * DISCLAIMER
 * Do not edit or add to this file if you wish to upgrade Bizuno to newer
 * versions in the future. If you wish to customize Bizuno for your
 * needs please refer to http://www.phreesoft.com for more information.
 *
 * @name       Bizuno ERP
 * @author     Dave Premo, PhreeSoft <support@phreesoft.com>
 * @copyright  2008-2022, PhreeSoft, Inc.
 * @license    http://opensource.org/licenses/OSL-3.0 Open Software License (OSL 3.0)
 * @version    6.x Last Update: 2021-11-15
 * @filesource /controllers/bizuno/export.php
 */

namespace bizuno;

class bizunoExport
{
    public  $moduleID  = 'bizuno';
    public  $user_email= '';      // will be filled in after access granted
    private $updated   = false;
    private $endCatOnly= false;   // if set to true, will only check the final category of the tree

    function __construct() {
        $this->lang = getLang('bizuno'); // needs to be hardcoded as this is extended by extensions
    }

    /**
     * Fetches the list of SKUs to upload
     * @param array $result - list of SKUs build upload list from
     * @return modified $layout
     */
    public function apiInvCount(&$layout=[], $result=[])
    {
        $output = [];
        foreach ($result as $row) { $output[] = $row['id']; }
        $layout = array_replace_recursive($layout, ['content' => ['items' => $output]]);
    }

    /**
     * Creates an inventory feed according to the API spec. Also allows for customization
     * @param array $product - array of filtered product data indexed by the product Tag name
     * @param string $url - remote site to send data
     * @param string $user - remote site username
     * @param string $pass - remote site password
     * @return array $data - structure of what to do next
     */
    public function apiInventory(&$product=[])
    {
        $rID = clean($product['RecordID'], 'integer');
        if (!$rID) { return msgDebug("\nBad ID passed. Needs to be the inventory field id tag name (RecordID)."); }
        // Extract the sales tax
        if (isset($product['TaxRateIDCustomer']) && $product['TaxRateIDCustomer'] > 0) {
            $rate = dbGetRow(BIZUNO_DB_PREFIX."tax_rates", "id={$product['TaxRateIDCustomer']}");
            $product['TaxRateTitle']   = $rate['title'];
            $product['TaxRatePercent'] = $rate['tax_rate'];
        } // else leave blank and use cart default tax setting
        unset($product['TaxRateIDCustomer']);
        unset($product['TaxRateIDVendor']);
        // load prices
        bizAutoLoad(BIZBOOKS_ROOT."controllers/inventory/prices.php", 'inventoryPrices');
        $prices = new inventoryPrices();
        $_GET['rID'] = $rID;
        $pDetails = [];
        $prices->quote($pDetails);
//      msgDebug("\nRetrieved price data in api = ".print_r($pDetails, true));
        $product['Price'] = $pDetails['content']['price'];
        if (!empty($pDetails['content']['regular_price'])){ $product['RegularPrice']= $pDetails['content']['regular_price']; }
        if (!empty($pDetails['content']['sale_price']))   { $product['SalePrice']   = $pDetails['content']['sale_price']; }
        if ( isset($pDetails['content']['sheets']) && sizeof($pDetails['content']['sheets']) > 0) { $product['PriceLevels'] = $pDetails['content']['sheets']; }
        $product['WeightUOM']   = getModuleCache('inventory', 'settings', 'general', 'weight_uom','LB');
        $product['DimensionUOM']= getModuleCache('inventory', 'settings', 'general', 'dim_uom',   'IN');
        switch ($product['sendMode']) {
            default:
            case 1:
            case 2:
                $this->getAccessories($product);
                $this->getImage($product);
                break;
            case 3:
            case 4: unset($product['Image']); break;
        }
    }

    /**
     * Sets the image path and populates image data, if present
     * @param type $product
     * @return modified $product
     */
    private function getImage(&$product)
    {
        global $io;
        $product['Images'] = [];
        $product['ProductImageData'] = $product['ProductImageDirectory'] = $product['ProductImageFilename'] = '';
        if (!empty($product['Image'])) { // primary image file
            $info = pathinfo($product['Image']);
            $data = $io->fileRead("images/{$product['Image']}");
            $product['ProductImageData']     = $data ? base64_encode($data['data']) : '';
            $product['ProductImageDirectory']= $info['dirname']."/";
            $product['ProductImageFilename'] = $info['basename'];
        }
        if (empty($product['invImages'])) { return; }
        $images = json_decode($product['invImages'], true);
        foreach ($images as $image) { // invImages extension list
            $info = pathinfo($image);
            $data = $io->fileRead("images/$image");
            $product['Images'][] = ['Title'=>$info['basename'], 'Path'=>$info['dirname']."/", 'Filename'=>$info['basename'], 'Data'=>$data ? base64_encode($data['data']) : ''];
        }
    }

    /**
     * Sets the accessory fields used when uploading inventory
     * @param array $product - working product data
     * @return array - modified $product
     */
    private function getAccessories(&$product)
    {
        if (isset($product['invAccessory'])) {
            $vals = json_decode($product['invAccessory'], true);
            if (!is_array($vals)) { return; }
            unset($product['invAccessory']);
            foreach ($vals as $rID) {
                $product['invAccessory'][] = dbGetValue(BIZUNO_DB_PREFIX."inventory", 'sku', "id=$rID");
            }
        }
    }

    /**
     * Method to sync product listings in Bizuno with shopping carts
     * @param array $layout -  current working structure
     * @param string $url - URL to send request to
     * @param string $user - user name at the cart
     * @param string $pass - password at the cart
     * @param string $field - field in Bizuno inventory table to look for cart enabled products
     * @param boolean $match - [default false] flag to delete at the cart, if present at cart and no flagged in Bizuno to sell
     */
    public function apiSync(&$output)
    {
        $output['syncDelete'] = clean('syncDelete', 'integer', 'get');
        $field  = !empty($output['syncTag']) ? $output['syncTag'] : 'cart_sync';
        $result = dbGetMulti(BIZUNO_DB_PREFIX.'inventory', "`$field`='1' AND inactive='0'");
        foreach ($result as $row) { $output['syncSkus'][] = $row['sku']; }
        msgDebug("\nSending aipSync content = ".print_r($output, true));
    }

    /**
     * This method syncs the products flagged in Bizuno to be listed and the actual listed products.
     * If syncDelete flag is set, the products in the cart will be deleted if they are not on the Bizuno list
     * @return messageStack entries
     */
    public function uploadSync($output)
    {
        $missing_skus = [];
        $cartProducts = wpdbGetMulti(PORTAL_DB_PREFIX.'postmeta', "meta_key='_sku'", 'meta_key');
        foreach ($cartProducts as $product_db) {
            if (!in_array($product_db['meta_value'], $output['syncSkus'])) {
                msgDebug("\nMissing product meta: ".print_r($product_db, true));
                $title = wpdbGetValue(PORTAL_DB_PREFIX.'posts', 'post_title', "ID='{$product_db['post_id']}'");
                $missing_skus[] = $title.' (SKU: '.$product_db['meta_value'].')';
                if (!empty($output['syncDelete'])) { wp_delete_post((int)$product_db['post_id']); }
            }
        }
        if (sizeof($missing_skus) > 0) { return msgAdd($this->lang['msg_sku_missing'].'Bizuno:'.'<br>'.implode('<br>', $missing_skus), 'caution'); }
        msgAdd($this->lang['msg_sku_sync_success'], 'success');
    }

    /**
     * Sends to ship confirmation status from Bizuno to the cart to update order status
     * @param array $data - working structure
     */
    public function apiConfirm(&$output)
    {
        $date = clean('dateShip', ['format'=>'date', 'default'=>biz_date('Y-m-d')], 'get');
        if (empty($date)) { $date = biz_date('Y-m-d'); }
        $this->confirmShip($output, $date);
        if (empty($output)) { msgAdd('No shipments could be found for this date!', 'success'); }

    }

    /**
     * Looks for product and replaces or creates a new product
     * @param array $product - data array of the product information
     * @return null - messages are added with status
     */
    public function uploadConfirm($orders=[])
    {
        $order_cnt = 0;
        $order_list= [];
        $prefix    = $this->configGet('bizuno_api_prefix');
        $status    = $this->configGet('bizuno_ship_status');
        $newStatus = $status ? $status : 'wc-completed';
        foreach ($orders as $id => $value) {
            msgDebug("\nWorking with id = $id and value = ".print_r($value, true));
            // strip prefix from order ref (WC3044 => 3044) that is the record number
            if     ($prefix && (strpos($id, $prefix) !== 0 || strpos($id, $prefix) === false)) { continue; }
            elseif ($prefix && strpos($id, $prefix) === 0) { $id = substr($id, strlen($prefix)); }
            $result = wpdbGetValue(PORTAL_DB_PREFIX.'posts', ['post_status','comment_status','comment_count'], "ID=".intval($id)." AND post_type='shop_order'");
            // if not found OR wp-completed continue - no need to duplicate
            if (empty($result) || $result['post_status'] == $newStatus) { msgDebug("\nSkipping id=$id"); continue; }
            // change the status to (wp-completed, closed, bump comment_count)
            $data = ['post_status'=>$newStatus,'comment_status'=>'closed','comment_count' =>$result['comment_count']+1];
            wpdbWrite(PORTAL_DB_PREFIX.'posts', $data, 'update', "ID='$id'");
            // insert into wp_comments per table fields with tracking, ship date etc.
            $comment = [
                'comment_post_ID'     => $id,
                'comment_author'      => 'Bizuno_Interfce',
                'comment_author_email'=> $this->user_email,
                'comment_date'        => wp_date('Y-m-d H:i:s'),
                'comment_date_gmt'    => gmdate('Y-m-d H:i:s'),
                'comment_content'     => $value['Tracking'],
                'comment_agent'       => 'WooCommerce',
                'comment_type'        => 'order_note',
            ];
            wpdbWrite(PORTAL_DB_PREFIX.'comments', $comment);
            $order_cnt++;
            $order_list[] = $id;
        }
        msgAdd(sprintf($this->lang['confirm_success'], sizeof($order_list), sizeof($order_list)>0?" (".implode(', ', $order_list).")":''), 'success');
        msgLog(sprintf($this->lang['ship_confirm_log'], $date));
    }

    /**
     * This method handles orders that have shipped on the specified date
     * @param array $output - working list of orders and shipping information
     * @param string $confirmDate - date to use as the ship date of the order
     */
    private function confirmShip(&$output, $confirmDate=false)
    {
        if (!$confirmDate) { return; }
        if (!dbTableExists(BIZUNO_DB_PREFIX.'extShipping')) {
            return msgAdd("The shipping extension is required for this operation.");
        }
        $result = dbGetMulti(BIZUNO_DB_PREFIX.'extShipping', "ship_date LIKE '$confirmDate%'");
        if (sizeof($result) == 0) { return msgAdd($this->lang['msg_no_shipments_found'], 'caution'); }
        // process the ship list
        foreach ($result as $value) {
            $invoice_num = strpos($value['ref_id'], '-') ? substr($value['ref_id'], 0, strpos($value['ref_id'], '-')) : $value['ref_id'];
            $value['order_num'] = dbGetValue(BIZUNO_DB_PREFIX.'journal_main', 'purch_order_id', "invoice_num='$invoice_num'");
            $order_num = $value['order_num'] ? $value['order_num'] : $invoice_num;
            $method_text = viewProcess($value['method_code'], 'shipInfo');
            if (isset($output[$order_num]['tracking_id'])) {
                $output[$order_num]['Tracking'] .= ', '.$value['tracking_id'];
            } else {
                $output[$order_num]['Tracking'] = 'Your order shipped '.viewDate($confirmDate).' via '.$method_text.', tracking number(s): '.$value['tracking_id'];
            }
            if (isset($output[$order_num]['notes'])) {
                $output[$order_num]['Notes'] .= '. '.$value['notes'];
            } else {
                if ($value['notes']) { $output[$order_num]['Notes'] = lang('notes').': '.$value['notes']; }
            }
            $output[$order_num]['Method']   = $method_text;
            $output[$order_num]['TrackID']  = $value['tracking_id'];
            $output[$order_num]['Status']   = 'orderShipped';
            $output[$order_num]['ShipDate'] = $confirmDate;
        }
    }

    /**
     * Looks for product and replaces or creates a new product
     * @param array $product - data array of the product information
     * @return null - messages are added with status
     */
    public function uploadProduct($product=[])
    {
        set_time_limit(240); // if uploading large images for firs time, it can take a while...
        if (empty($product)) { msgAdd("Bad product string received, check the error log."); }
        msgDebug("\nReached uploadProduct with size of product array = ".sizeof($product));
        if (empty($product['SKU'])) { return msgAdd("Cannot find product SKU."); }
        $this->productImport($product);
    }

    /**
     * Starts the import of product to WooCommerce
     * @param type $product
     * @return type
     */
    private function productImport($product)
    {
        $user_id = get_current_user_id();
        $slug = !empty($product['WooCommerceSlug']) ? $product['WooCommerceSlug'] : $product['WooCommerceCategory']."_".$product['Model'];
        $product_post_array = [
            'post_content'  => !empty($product['DescriptionLong']) ? $product['DescriptionLong'] : $product['DescriptionSales'],
            'post_title'    => $product['Description'],
            'post_excerpt'  => !empty($product['DescriptionSales'])? $product['DescriptionSales'] : $product['Description'],
            'post_name'     => $this->getPermaLink($slug),
            'menu_order'    => !empty($product['MenuOrder']) ? (int)$product['MenuOrder'] : 99];
        $product_meta_array = [
            '_downloadable' => !empty($product['Downloadable'])    ? 'yes': 'no',
            '_manage_stock' => !empty($product['Virtual'])         ? 'no' : 'yes',
            '_price'        => !empty($product['SalePrice'])       ? $product['SalePrice']    : $product['Price'],
            '_regular_price'=> !empty($product['RegularPrice'])    ? $product['RegularPrice'] : $product['Price'],
            '_stock'        => $product['QtyStock'] > 0            ? $product['QtyStock']     : 0,
            '_stock_status' => $product['QtyStock'] > 0            ? 'instock' : 'outofstock',
            '_tax_class'    => '',
            '_tax_status'   => 'taxable',
            '_virtual'      => !empty($product['Virtual'])         ? 'yes' : 'no'];
        //covert weight (need to convert kg,lb,oz,g)
        if (isset($product['WeightUOM'])) {
            $weight = strtolower($product['WeightUOM']);
            $wp_weight = strtolower($this->configGet('woocommerce_weight_unit'));
            $product_meta_array['_weight'] = isset($product['Weight']) ? $this->convertWeight($product['Weight'], $weight, $wp_weight) : '';
        }
        //covert dim (need to convert m,cm,mm,in,yd)
        if (isset($product['DimensionUOM'])) {
            $dim = strtolower($product['DimensionUOM']);
            $wordpress_dim = strtolower($this->configGet('woocommerce_dimension_unit'));
            $product_meta_array['_length']= isset($product['ProductLength'])? $this->convertLength($product['ProductLength'],$dim, $wordpress_dim) : '';
            $product_meta_array['_width'] = isset($product['ProductWidth']) ? $this->convertLength($product['ProductWidth'], $dim, $wordpress_dim) : '';
            $product_meta_array['_height']= isset($product['ProductHeight'])? $this->convertLength($product['ProductHeight'],$dim, $wordpress_dim) : '';
        }
        //post the product
        $productID = $this->getProductID($product['SKU']);
        if (empty($productID)) { // insert new product
            // build the products table data
            $product_post_new = [
                'post_author'   => $user_id, //id
                'post_date'     => !empty($product['DateCreated']) ? $product['DateCreated'] : wp_date('Y-m-d H:i:s'),
                'post_date_gmt' => !empty($product['DateCreated']) ? $product['DateCreated'] : wp_date('Y-m-d H:i:s'),
                'post_status'   => 'publish', // allows the product to be displayed (default)
                'comment_status'=> 'open',
                'ping_status'   => 'closed',
                'post_type'     => 'product'];
            $product_meta_new = [
                '_sku'          => $product['SKU'],
                '_visibility'   => 'visible'];
            $postData = array_merge($product_post_array, $product_post_new);
            $postMeta = array_merge($product_meta_array, $product_meta_new);
            $product_id = $this->setPost($postData, $postMeta, $productID);
            if (empty($product_id)) { return msgDebug("\nError posting the WordPress way, no ID returned!"); }
        } else { $product_id = $productID; }
        if (empty($product['SalePrice'])) { delete_post_meta($product_id, '_sale_price'); }
        else                              { update_post_meta($product_id, '_sale_price', $product['SalePrice']); }
        switch ($product['sendMode']) {
            default:
            case 1: $replaceImage = true;// Full Upload (Slowest - replace/regenerate all images)
            case 2: // Full Product Details (Skip images if present)
                $this->productImage($product, $product_id, !empty($replaceImage) ? true : false); // Set images
            case 3: // Product Core Info (No Categories/Images)
                $this->productAttributes($product, $product_id); // Update attributes
                $this->productRelated($product); // Set related products
                $this->productVariations($product, $product_id); // check for master stock type
                $this->productMetadata($product, $product_id);
                $this->productTags($product, $product_id);
                $this->productCategory($product, $product_id); //update category
            case 4: // Update pricing/stock only (Fastest)
                $this->productData($product_post_array, $product_meta_array, $product_id);
                $this->productPrices($product, $product_id); // update the price levels
                break;
        }
        msgAdd("Successfully imported SKU: {$product['SKU']}", 'success');
        return $product_id;
    }

    private function productData($post_array, $meta_array, $pID)
    {
        wpdbWrite(PORTAL_DB_PREFIX.'posts', $post_array, 'update', "ID=$pID");
        foreach ($meta_array as $key => $value) {
//          update_post_meta($product_id, $key, $value);  // the right way but may be slower
            wpdbWrite(PORTAL_DB_PREFIX.'postmeta', ['meta_value'=>$value], 'update', "post_id=$pID AND meta_key='$key'");
        }
    }

    private function productRelated($product)
    {
        // fetch related id
        if (!empty($product['invAccessory']) && is_array($product['invAccessory'])) {
            $product['related'] = [];
            foreach ($product['invAccessory'] as $related) {
                $product_id = wpdbGetValue(PORTAL_DB_PREFIX.'postmeta', 'post_id', "`meta_key` LIKE '_sku' AND `meta_value`='{$related}'", true);
                if ($product_id !== false) { $product['related'][] = $product_id; }
            }
            msgDebug("related items found:".print_r($product['related'], true));
        }
        if (isset($product['related'])) {
//            wpdbGetResult("DELETE FROM `". PORTAL_DB_PREFIX . "postmeta` WHERE post_id = '". (int)$product_id . "' AND meta_key = '_crosssell_ids';");
//            wpdbGetResult("INSERT INTO " . PORTAL_DB_PREFIX . "postmeta SET post_id = '"   . (int)$product_id . "', meta_key = '_crosssell_ids' , meta_value = '" . $product['related'] . "';");
        }
    }

    /**
     *
     * @param type $sku
     * @return int
     */
    protected function getProductID($sku) {
        $candidates = wpdbGetMulti(PORTAL_DB_PREFIX.'postmeta', "`meta_key`='_sku' AND `meta_value`='$sku'", '', ['post_id']);
        msgDebug("\nSearching for existing product, $sku and found: ".print_r($candidates, true));
        $firstID = 0;
        foreach ($candidates as $val) {
            $postID = wpdbGetValue(PORTAL_DB_PREFIX.'posts', 'ID', "`ID`='{$val['post_id']}' AND `post_status`='publish'");
            if ($postID) {
                if (empty($firstID)) {
                    msgDebug("\nSetting first ID = $postID");
                    $firstID =  $postID;
                } else { // duplicate published products, delete it
                    wp_delete_post( $postID, true );
                }
            }
        }
        return $firstID;
    }

    /**
     * Enters the volume price if included with feed
     *
     * REQUIRES THE WooCommerce Bulk Pricing plug-in set to percentage discount (default)
     * https://wordpress.org/plugins/woocommerce-bulk-discount/
     *
     * @param type $product
     * @param type $product_id
     * @return type
     */
    protected function productPrices($product, $product_id)
    {
        msgDebug("\nEntering productPrices and checking for Woocommerce Bulk Discount plugin active");
        if ( !is_plugin_active( 'woocommerce-bulk-discount/woocommerce-bulk-discount.php' ) ) { return; }
        $resetMeta = true;
        if (!empty($product['PriceLevels']) && is_array($product['PriceLevels']) ) {
            foreach ($product['PriceLevels'] as $price) {
                if (empty($price['default'])) { continue; } // only use values from default price sheet, if specified
                $description = '<table style="border:1px solid black;width:300px"><tr><th colspan="2">Quantity Pricing</th></tr><tr><th>Qty</th><th>Price</th></tr>';
                for ($i=0; $i<sizeof($price['levels']); $i++) {
                    if ($i==0) { // first price is single qty, save to calc discount
                        $full_price = !empty($price['levels'][$i]['price']) ? $price['levels'][$i]['price'] : 999999;
                        continue;
                    } // skip the first price level as it is the single unit price
                    $resetMeta = false; // if we are here, there is more than 1 price in this discount group
                    $description .= '<tr><td style="border: 1px solid black;text-align:center">'.$price['levels'][$i]['qty'].'</td><td style="border: 1px solid black;text-align:center">$ '.number_format($price['levels'][$i]['price'], 2).'</td></tr>';
                    msgDebug("\nUpdating pricing for quantity: ".$price['levels'][$i]['qty']);
                    update_post_meta($product_id, "_bulkdiscount_quantity_$i",      $price['levels'][$i]['qty']);
                    update_post_meta($product_id, "_bulkdiscount_discount_$i",      round(((1 - $price['levels'][$i]['price']/$full_price) * 100), 3));
                    update_post_meta($product_id, "_bulkdiscount_discount_fixed_$i",round(($full_price - $price['levels'][$i]['price']), 3));
                }
                // clean up extra levels, if they are present, i.e. number of discount levels were reduced.
                for ($j=$i+1; $j<100; $j++) {
                    $idx = get_post_meta( $product_id, "_bulkdiscount_quantity_$j");
                    if (empty($idx)) {
                        $j=100; // stop the loop
                    } else {
                        delete_post_meta( $product_id, "_bulkdiscount_quantity_$j");
                        delete_post_meta( $product_id, "_bulkdiscount_discount_$j");
                        delete_post_meta( $product_id, "_bulkdiscount_discount_fixed_$j");
                    }
                }
            }
            $description .= '</table>';
        }
        msgDebug("\nSetting the meta data for the description and enable flag");
        if ($resetMeta) { // clear any fields that may have been there from a prior discount
            update_post_meta( $product_id, '_bulkdiscount_enabled', 'no');
            delete_post_meta( $product_id, '_bulkdiscount_text_info');
        } else {
            update_post_meta( $product_id, '_bulkdiscount_enabled', 'yes');
            update_post_meta( $product_id, '_bulkdiscount_text_info', $description);
        }
    }

    private function productMetadata($product, $product_id)
    {
        msgDebug("\nEntering productMetadata and checking for YOST SEO plugin active");
        if ( !is_plugin_active( 'wordpress-seo/wp-seo.php' ) ) { return; }
        if (!empty($product['MetaDescription'])) {
            update_post_meta( $product_id, '_yoast_wpseo_metadesc', $product['MetaDescription']);
        }
        if (!empty($product['SearchCode'])) {
            update_post_meta( $product_id, 'biz_search_code', $product['SearchCode']);
        }
    }

    /**
     * Set the tags
     * @param type $product
     * @param type $product_id
     * @return boolean
     */
    private function productTags($product, $product_id)
    {
        if (empty($product['WooCommerceTags'])) { return; }
        msgDebug("\nWorking with raw tags: {$product['WooCommerceTags']}");
        $tags = explode(";", $product['WooCommerceTags']);
        wp_set_object_terms($product_id, $tags, 'product_tag');
        return true;
    }

    /**
     *
     * @param type $product
     * @param type $product_id
     * @return boolean
     */
    private function productCategory($product, $product_id)
    {
        if (empty($product['WooCommerceCategory'])) {
            return msgAdd("Error - the category was not passed for product: {$product['SKU']}, it must be set manually in WooCommerce.", 'caution');
        }
        msgDebug("\nWorking with raw category: {$product['WooCommerceCategory']}");
        $categories = explode(";", $product['WooCommerceCategory']);
        foreach ($categories as $category) {
            $parent = 0;
            $descName = $niceName = '';
            $breadcrumbs = explode(":", $category);
            foreach ($breadcrumbs as $breadcrumb) { // starts at the top level and goes down.
                $term = trim($breadcrumb);
                if (empty($term)) { continue; }
                $descName = trim($term);
//              $descName = trim($descName.' '.$term);
                $niceName = trim(strtolower(str_replace([':',' '], '-', $term)), " -");
//              $niceName = trim(strtolower(str_replace([':',' '], '-', $niceName.' '.$term)), " -");
                $termIDs  = term_exists( $descName, 'product_cat', !empty($parent)?$parent:null );
                msgDebug("\nSearching for descName = $descName, parent = $parent and resulted in termIDs = ".print_r($termIDs, true));
                if (empty($termIDs)) {
                    $termData= [ 'slug'=>$niceName, 'parent'=>$parent ]; // 'description'=>$descName, // leave description blank so user can edit through WooCommerce
                    msgDebug("\nInserting term data: ".print_r($termData, true));
                    $termIDs = wp_insert_term( $descName, 'product_cat', $termData );
                }
                msgDebug("\nSetting post termIDs: ".print_r($termIDs, true));
                wp_set_post_terms( $product_id, [$termIDs['term_id']], 'product_cat', $this->endCatOnly?false:true );
                $parent = $termIDs['term_id'];
            }
            // Check just the lowest on the category tree if endCatOnly is true
            if ($this->endCatOnly) { wp_set_post_terms( $product_id, [$termIDs['term_id']], 'product_cat', true ); }
        }
        return true;
    }

    /**
     * Add product attributes, this method just hard codes the value and avoids terms and taxonomy which creates a new term for every possible value
     * @param type $product
     * @param type $product_id
     * @return type
     */
    private function productAttributes($product, $product_id)
    {
        if (empty($product['Attributes'])) { return; }
        $result = wpdbGetMulti(PORTAL_DB_PREFIX.'term_taxonomy', "taxonomy LIKE 'pa_%'");
        $pa_attr_ids = [];
        foreach ($result as $row) { $pa_attr_ids[] = $row['term_taxonomy_id']; }
        if (sizeof($pa_attr_ids)) { // clear out the current attributes
            wpdbGetResult("DELETE FROM ".PORTAL_DB_PREFIX."term_relationships WHERE object_id=$product_id AND term_taxonomy_id IN (".implode(',',$pa_attr_ids).")");
        }
        $productAttr = [];
        foreach ($product['Attributes'] as $idx => $row) {
            if (empty($row['title']) || empty($row['index'])) { continue; }
            $attrSlug= $this->getPermaLink($row['index']);
//          $attrSlug= $this->getPermaLink($product['AttributeCategory'].'_'.$row['index']); // creates a lot of attributes and causes filtering issues
            $exists  = wpdbGetValue(PORTAL_DB_PREFIX.'woocommerce_attribute_taxonomies', 'attribute_name', "attribute_name='$attrSlug'");
            if (!$exists) {
                $newAttr = [
                    'attribute_name'   => $attrSlug,
                    'attribute_label'  => $row['title'],
                    'attribute_type'   => 'text',
                    'attribute_orderby'=> 'name_num',
                    'attribute_public' => 0];
                wpdbWrite(PORTAL_DB_PREFIX.'woocommerce_attribute_taxonomies', $newAttr);
            }
            $productAttr["pa_$attrSlug"] = ['name'=>$row['title'],'value'=>$row['value'],'position'=>$idx,'is_visible'=>1,'is_variation'=>0,'is_taxonomy'=>0];
            // Update postmeta with attribute key => value pair for searching...
            update_post_meta( $product_id, "biz_{$row['index']}", $row['value'] );
        }
        msgDebug("\nUpdating product postmeta: ".print_r($productAttr, true));
        update_post_meta( $product_id, '_product_attributes', $productAttr );
    }

    /**
     *
     * @param type $value
     * @return type
     */
    private function getPermaLink($value)
    {
        $test1 = str_replace(':', '_', $value);
        $test2 = str_replace([' ', '/', '.'], '-', trim($test1));
        $test3 = preg_replace("/[^a-zA-Z0-9\-\_]/", "", $test2);
        while (strpos($test3, '--') !== false) { $test3 = str_replace('--', '-', $test3); }
        return strtolower($test3);
    }

    /**
     *
     * @param type $product
     * @param type $pID
     * @return type
     */
    private function productVariations($product, $pID) {
        msgDebug("\nEntering productVariations with invOptions = ".print_r($product['invOptions'], true));
        if (empty($product['Type']) || empty($product['invOptions']) || $product['Type']<>'ms') { return; }
        $cnt     = 0;
        $skuList = [$product['SKU'].'-'];
        $options = $product['invOptions'];
        for ($i=0; $i<sizeof($options); $i++) {
            $slug = strtolower(str_replace(' ', '-', $options[$i]['option']));
            $output[$slug] = ['name'=>$options[$i]['option'], 'value'=>str_replace(';', ' | ', $options[$i]['labels']),
                'position'=>$cnt, 'is_visible'=>1, 'is_variation'=>1, 'is_taxonomy'=>0];
            $skuList = $this->BuildSKU($skuList, $options[$i]['attrs']);
            $cnt++;
        }
        // Set the attributes for the main SKU
        msgDebug("\nWriting attributes list = ".print_r($output, true));
        $meta['_product_attributes'] = serialize($output);
//      $meta['_default_attributes'] = serialize([]); // not used, can be set manually in Woo
        $this->setPost([], $meta, $pID);
        $term_id = wpdbGetValue(PORTAL_DB_PREFIX.'terms',         'term_id',       "name='variable'");
//      $taxy_id = wpdbGetValue(PORTAL_DB_PREFIX.'term_taxonomy', 'term_taxonomy', "taxonomy='product_type'"); // one to one with terms
        $isThere = wpdbGetValue(PORTAL_DB_PREFIX.'term_relationships', 'object_id', "object_id=$pID AND term_taxonomy_id=$term_id");
        if (!$isThere) {
            wpdbWrite(PORTAL_DB_PREFIX.'term_relationships', ['object_id'=>$pID,'term_taxonomy_id'=>$term_id]);
        }
        // find the sku's and set the parent
        msgDebug("\nLeaving productVariations with SKU list to find = ".print_r($skuList, true));
        foreach ($skuList as $sku) {
            $mID  = wpdbGetValue(PORTAL_DB_PREFIX.'postmeta', 'post_id', "`meta_key`='_sku' AND `meta_value`='$sku'");
            msgDebug("\nResult of sku search = ".print_r($mID, true));
            if (!empty($mID)) { wpdbWrite(PORTAL_DB_PREFIX.'posts', ['post_parent'=>$pID, 'post_type'=>'product_variation'], 'update', "ID='$mID'"); }
        }
    }

    /**
     *
     * @param type $list
     * @param type $attrs
     * @return string
     */
    private function BuildSKU($list, $attrs)
    {
        $output = [];
        $opts   = explode(';', $attrs);
        foreach ($list as $sku) {
            foreach ($opts as $value) { $output[] = $sku . $value; }
        }
        return $output;
    }

    /**
     *
     * @param type $product
     * @param type $product_id
     * @return type
     */
    private function productImage($product, $product_id, $replace=false)
    {
        msgDebug("\nEntering Image function with product ID = $product_id");
        $media = [];
        require_once( ABSPATH.'wp-admin/includes/image.php' );
        $this->setImageProps($media, $product['ProductImageDirectory'], $product['ProductImageFilename'], $product['ProductImageData']);
        msgDebug("\nReady to process extra Images with size of Images tag = ".sizeof($product['Images']));
        if (!empty($product['Images']) && is_array($product['Images'])) {
            foreach ($product['Images'] as $image) {
//                msgDebug("\nimage array = ".print_r($image, true));
                $this->setImageProps($media, $image['Path'], $image['Filename'], $image['Data']);
            }
        } else { msgDebug("\nOnly one image, it will become the thumbnail."); }
        if (empty($media)) {
            msgDebug("\nReturning from productImage with no images found!");
            return;
        } // No images uploaded
        $this->setImageCleaner($product_id); // takes out the trash
        // ready to set images, since they are searched and id'ed based on the path, we only need the meta index to retain the position
//      $metaIdx= wpdbGetValue(PORTAL_DB_PREFIX.'postmeta', 'meta_id', "post_id=$product_id AND meta_key='_thumbnail_id'");
        $props  = array_shift($media);
        $imgIdx = $this->setImage($props, $product_id, $replace);
        msgDebug("\nReturned from setImage with thumbnail post ID = $imgIdx");
        if (!empty($imgIdx)) { update_post_meta( $product_id, '_thumbnail_id', $imgIdx ); }
        // Set the image gallery (for the rest of the images)
        msgDebug("\nReady to process extra images with media = ".print_r($media, true));
        $xIDs   = [];
        foreach ($media as $props) {
            $imgIdx = $this->setImage($props, $product_id, $replace);
            if (!empty($imgIdx)) { $xIDs[] = $imgIdx; }
        }
        update_post_meta($product_id, '_product_image_gallery', implode(',', $xIDs));
    }

    private function setImageProps(&$media, $path='', $name='', $data='')
    {
        if (empty($data)) { return; }
        $tmp0 = 'products/'.(!empty($path) ? $path : ''); // from root upload folder
        $tmp1 = str_replace('/./', '/', $tmp0);
        $media[] = ['path' => rtrim($tmp1, '/').'/', 'name' => $name, 'data' => $data];
    }

    /**
     * Cleans out duplicates and other issues from earlier releases
     * @param type $activeIDs
     * @param type $product_id
     */
    private function setImageCleaner($product_id)
    {
        msgDebug("\nEntering setImageCleaner with product_id = $product_id");
        // first check thumbnails for multiple records, should only be one
        $metaIDs = wpdbGetMulti(PORTAL_DB_PREFIX.'postmeta', "post_id=$product_id AND meta_key='_thumbnail_id'");
        if (sizeof($metaIDs) > 1) {
            for ($i=1; $i<sizeof($metaIDs); $i++) { // earlier bug where multiple thumbnails were generated
                msgDebug("\nDeleting duplicate thumbnail with ID = ".print_r($metaIDs[$i], true));
                wpdbGetResult("DELETE FROM ".PORTAL_DB_PREFIX."postmeta WHERE meta_id={$metaIDs[$i]['meta_id']}");
                \wp_delete_post( $metaIDs[$i]['post_id'], true );
            }
        }
        // If the same image is used for multiple products, then multiple media posts were generated, clean these up and start over.
        $dupImages  = wpdbGetMulti(PORTAL_DB_PREFIX.'posts', "post_parent<>0 AND post_parent=$product_id AND post_type='attachment'", 'ID', ['ID']);
        foreach ($dupImages as $imageID) {
            msgDebug("\nDeleting duplicate images with ID = ".print_r($imageID, true));
            \wp_delete_post( $imageID['ID'], true );
        }
    }

    /**
     * Uploads the image and puts it in the proper folder, creates to folder if it can.
     * @param array $props
     * @param integer $imgID
     * @param integer $product_id
     * @return type
     */
    private function setImage($props, $product_id, $replace=false) {
        msgDebug("\nEntering setImage with props = ".print_r($props, true));
        $upload_folder= wp_upload_dir(); // wp default is "wp-content/uploads" - needs trailing slash
        $image_dir    = $upload_folder['basedir']."/{$props['path']}";
        $filename     = $image_dir.$props['name']; // '/path/to/uploads/2013/03/filename.jpg';
        $guid         = $props['path'] . $props['name'];
        msgDebug("\nLooking for all images at: $guid");
        // BOF - Clean out duplicate image posts pointing to the same file
        $postIDs = wpdbGetMulti(PORTAL_DB_PREFIX.'postmeta', "meta_key='_wp_attached_file' AND meta_value='$guid'", 'post_id', ['post_id']);
        msgDebug("\nRead the following ID's for this image path: ".print_r($postIDs, true));
        for ($i=1; $i<sizeof($postIDs); $i++) { // earlier bug where multiple thumbnails were generated pointing to same file location
            msgDebug("\nDeleting duplicate posts with same path and ID = {$postIDs[$i]['post_id']}");
            wpdbGetResult("DELETE FROM ".PORTAL_DB_PREFIX."posts WHERE ID={$postIDs[$i]['post_id']}");
            wpdbGetResult("DELETE FROM ".PORTAL_DB_PREFIX."postmeta WHERE post_id={$postIDs[$i]['post_id']}");
//            \wp_delete_post( $postIDs[$i]['post_id'], true ); // seemed to leave some orphan meta data
        }
        if (sizeof($postIDs)>0) {
            $postExists = wpdbGetValue(PORTAL_DB_PREFIX.'posts', 'ID', "ID={$postIDs[0]['post_id']}");
            msgDebug("\nRead to see if the record exists: ".print_r($postExists, true));
            if (empty($postExists)) { wpdbGetResult("DELETE FROM ".PORTAL_DB_PREFIX."postmeta WHERE post_id={$postIDs[0]['post_id']}"); }
            $imgID = !empty($postExists) ? $postIDs[0]['post_id'] :  0;
        } else {
            $imgID = 0;
        }
        // EOF - Clean out duplicate images
        msgDebug("\nWorking with image ID = $imgID. Now testing for image length = ".strlen($props['data']));
        if (!$props['data']) { return; } // no image was sent up to save, just return with no message
        // If skip overwrite and image is present, return with just the ID
        if (!$replace && !empty($imgID)) {
            msgDebug("\nReplace is set to false and the image exists, returning image ID = $imgID");
            return $imgID;
        }
        // NOTE: the str_replace is to necessary to fix a PHP 5 issue with spaces in the base64 encode... see php.net
        $contents    = base64_decode(str_replace(" ", "+", $props['data']));
        if (!is_dir($image_dir)) { if (!@mkdir($image_dir, 0755, true)) { return msgAdd("Cannot create image path: $image_dir"); } }
        $full_path   = $image_dir.$props['name'];
        $dirname     = dirname($full_path);
        if (!is_dir($dirname)) { mkdir($dirname, 0755, true); }
        if (!$handle = fopen($full_path, 'wb')) { return msgAdd("Cannot open Image path: $full_path"); }
        if (fwrite($handle, $contents) === false) { return msgAdd("Cannot write Image file."); }
        fclose($handle);
        msgDebug("\nWrote image image_directory = $image_dir and image_filename = {$props['name']} and image length = ".strlen($props['data']));
        $filetype = wp_check_filetype(basename( $filename ), null);
        $args = [
            'guid'          => $upload_folder['baseurl'] . "/$guid",
            'post_mime_type'=> $filetype['type'],
            'post_title'    => preg_replace( '/\.[^.]+$/', '', basename( $filename ) ),
            'post_content'  => '',
            'post_type'     => 'attachment',
            'post_status'   => 'inherit'];
        if (!empty($imgID)) { $args['ID'] = $imgID; }
        msgDebug("\nReady to insert image from $filename with args = ".print_r($args, true));
        $attach_id = \wp_insert_post( $args, true );
        msgDebug("\nFinished inserting Image with returned id = ".print_r($attach_id, true));
        if (!is_wp_error($attach_id)) {
            if ($attach_id==0) { msgDebug("\nForced attachment_id to be = $imgID"); $attach_id = $imgID; } // for some reason, WP returns 0 when the ID is set going into post and no error
            update_post_meta( $attach_id, '_wp_attached_file', $guid ); // this needs to be there at a minimum or media details will not render image
            update_post_meta( $attach_id, '_wp_attachment_metadata', ['file'=>$filename] ); // this needs to be there at a minimum or media details will not render image
            $fileParent = $this->getFBParent($props['path']);
            if (false !== $fileParent) { $this->setFBAttach($attach_id, $fileParent); }
            $attach_data = wp_generate_attachment_metadata( $attach_id, $filename );
            msgDebug("\nFinished wp_generate_attachment_metadata");
            wp_update_attachment_metadata( $attach_id, $attach_data ); // TAKES REALLY LONG, UP TO A MINUTE, MOVE TO CRON
            msgDebug("\nfinshed wp_update_attachment_metadata");
            msgDebug("\nReturning from setImage, set image ID = $attach_id");
            return $attach_id;
        }
        msgDebug("\nReturning from setImage with error!");
        return false;
    }

    /**
     *
     * @param type $entry
     * @return string
     */
    private function getImageType($entry)
    {
        $ext = strtolower(substr($entry, strrpos($entry, '.')+1));
        if (in_array($ext, ['png'])) { return 'image/png'; }
        if (in_array($ext, ['jpg','jpeg'])) { return 'image/jpeg'; }
        if (in_array($ext, ['gif'])) { return 'image/gif'; }
        return '';
    }

    private function getFBParent($path)
    {
        global $wpdb;
        msgDebug("\nEntering getFBParent with path = $path");
        if ( is_plugin_active( 'filebird/filebird.php' ) || is_plugin_active( 'filebird-pro/filebird.php' ) ) {
            $clnPath= rtrim(trim($path), '/');
            msgDebug("\nCleaned path = $clnPath");
            if (empty($clnPath)) { return false; }
            $dirs   = explode("/", $clnPath);
            msgDebug("\nEXploded into dirs = ".print_r($dirs, true));
            $parent = 0;
            foreach ($dirs as $dir) {
                $sql = "SELECT id FROM {$wpdb->prefix}fbv WHERE name='$dir' AND parent=$parent";
                $result = $wpdb->get_row($sql);
                if (is_null($result)) {
                    msgDebug("\nInserting into fbv with dir = $dir and parent = $parent");
                    wpdbGetResult($wpdb->prepare("INSERT INTO {$wpdb->prefix}fbv (name, parent) VALUES ('$dir', $parent)"));
                    $parent = $wpdb->insert_id;
                } else {
                    $parent = $result->id;
                }
            }
            msgDebug("\nreturning from getFBParent with parent: $parent");
            return $parent;
        }
    }

    private function setFBAttach($attach_id, $fileParent)
    {
        global $wpdb;
        $result = $wpdb->get_row("SELECT folder_id FROM {$wpdb->prefix}fbv_attachment_folder WHERE folder_id=$fileParent AND attachment_id=$attach_id");
        msgDebug("\nRead from fbv_attachment_folder: ".print_r($result, true));
        if (is_null($result)) {
            msgDebug("\nInserting into fbv_attachment_folder parent = $fileParent and attach_id = $attach_id");
            wpdbGetResult($wpdb->prepare("INSERT INTO {$wpdb->prefix}fbv_attachment_folder (folder_id, attachment_id) VALUES ($fileParent, $attach_id)"));
        }
    }

    /**
     * Inserts/updates the postmeta for a specified post_id
     * @param array $postData - for the post main record
     * @param array $postMeta - for the post meta table
     * @param integer $post_id - if known, database record, if empty then a new record will be created
     * @return integer - post id, relevant if a new record was created
     */
    protected function setPost($postData=[], $postMeta=[], $post_id=0)
    {
        $postData['ID'] = $post_id;
        $data = array_merge($postData, ['meta_input'=>$postMeta]);
        msgDebug("\nInserting/Updating db table posts/postmeta with data = ".print_r($data, true));
        $postID = wp_insert_post($data, true);
        if (is_wp_error($postID)) {
            $errors = $postID->get_error_messages();
            msgDebug("WP DB update error: ".print_r($errors, true));
            return 0;
        }
        return $postID;
    }

    /**
     * Takes package information in and returns the Bizuno enabled shipping rates and services
     * @param array $layout - Structure coming in ($layout['pkg'] is where the information is)
     */
    public function shippingRates(&$layout=[])
    {
        if ( !\is_plugin_active( 'bizuno-pro/bizuno-pro.php' ) ) { return; }
        $pkg = [
            'ship' => [
                'address1'   => $layout['pkg']['destination']['address'],
                'address2'   => $layout['pkg']['destination']['address_1'],
                'city'       => $layout['pkg']['destination']['city'],
                'state'      => $layout['pkg']['destination']['state'],
                'postal_code'=> $layout['pkg']['destination']['postcode'],
                'country'    => clean($layout['pkg']['destination']['country'], ['format'=>'country','option'=>'ISO3'])],
            'settings' => [
                'ship_date'  => date('Y-m-d'),
                'order_total'=> $layout['pkg']['cart_subtotal'],
                'weight'     => !empty($layout['pkg']['destination']['totalWeight']) ? $layout['pkg']['destination']['totalWeight'] : 1,
//              'length'     => clean('length',     'float',  'post'),
//              'width'      => clean('width',      'float',  'post'),
//              'height'     => clean('height',     'float',  'post'),
                'num_boxes'  => 1, // Assume 1 for now as dims may not be entered
                'ltl_class'  => 60, // clean('ltl_class',  'text',   'post'),
                'residential'=> 1,  // clean('residential','boolean','post'),
                'verify_add' => true]];
        $pkg['ship']['country2'] = $layout['pkg']['destination']['country'];
        $pkg['ship']['country3'] = $pkg['ship']['country'];
        require_once(BIZBOOKS_EXT.'proLgstc/ship.php');
        $quote = new proLgstcShip();
        $quote->getRateAddress($pkg, 'bill');
        $quote->getRateAddress($pkg, 'origin');
        $layout['rates'] = $quote->rateAPI($pkg);
    }

    /**
     *
     * @param unknown $value
     * @param unknown $unit_original m,cm,mm,in,ft
     * @param unknown $unit_return m,cm,mm,in,yd
     */
    private function convertLength($value, $unit_original, $unit_return) {
        if($unit_original == $unit_return) { return $value; }
        switch ($unit_original) {
            case 'm':
                switch($unit_return) {
                    case 'cm': return round($value/100, 2);
                    case 'mm': return round($value/1000, 2);
                    case 'in': return round($value*39.370, 2);
                    case 'yd': return round($value/1.0936, 2);
                }
            case 'cm':
                switch($unit_return) {
                    case  'm': return round($value*100, 2);
                    case 'mm': return round($value/10, 2);
                    case 'in': return round($value*0.39370, 2);
                    case 'yd': return round($value/109.36, 2);
                }
            case 'mm':
                switch($unit_return) {
                    case  'm': return round($value*1000, 2);
                    case 'cm': return round($value*10, 2);
                    case 'in': return round($value*0.039370, 2);
                    case 'yd': return round($value*0.039370*3*12, 2);
                }
            case 'in':
                switch($unit_return) {
                    case  'm': return round($value/39.370, 2);
                    case 'cm': return round($value/0.39370, 2);
                    case 'mm': return round($value/0.039370, 2);
                    case 'yd': return round($value*3*12, 2);
                }
            case 'ft':
                switch($unit_return) {
                    case  'm': return round($value/3.2808, 2);
                    case 'cm': return round($value/0.032808, 2);
                    case 'mm': return round($value/0.0032808, 2);
                    case 'yd': return round($value*3, 2);
                    case 'in': return round($value/12, 2);
                }
            default:
                msgAdd("length conversion Error","warning");
                return $value;
        }
    }

    /**
     *
     * @param unknown $value
     * @param unknown $unit_original kg,lb
     * @param unknown $unit_return kg,g,lb,oz
     * @return void|unknown|string|number
     */
    private function convertWeight($value, $unit_original, $unit_return) {
        msgDebug("\nEntering convertWeight with value = $value, unit_original = $unit_original and unit_return = $unit_return");
        switch ($unit_original) {
            case 'kg':
            case 'kgs':
                if (in_array($unit_original, ['kg', 'kgs'])) { return value; }
                switch($unit_return) {
                    case  'g': return round($value/1000, 3);
                    case 'lb':
                    case 'lbs':return round($value*2.2046, 2);
                    case 'oz': return round($value*35.274, 2);
                }
            case 'lb':
            case 'lbs':
                if (in_array($unit_original, ['lb', 'lbs'])) { return $value; }
                switch($unit_return) {
                    case 'kg':
                    case 'kgs':return round($value/2.2046, 3);
                    case  'g': return round($value/0.0022046, 3);
                    case 'oz': return round($value*16, 2);
                }
            default:
                msgAdd("Weight Conversion Error, received original unit = $unit_original", 'warning');
                return $value;
        }
    }

    /**
     * Pulls an WP option or array of options
     * @param type $const
     * @return type
     */
    private function configGet($const='') {
        if (is_array($const)) {
            $output = [];
            foreach ($const as $value) { $output[$value] = get_option($value); }
        } else {
            $output = get_option($const);
        }
        return $output;
    }

    /**
     * Install common fields into inventory db table shared amongst all the interfaces
     */
    protected function installStoreFields()
    {
        $id = validateTab($module_id='inventory', 'inventory', lang('estore'), 80);
        if (!dbFieldExists(BIZUNO_DB_PREFIX."inventory", 'msrp'))             { dbGetResult("ALTER TABLE ".BIZUNO_DB_PREFIX."inventory ADD msrp DOUBLE NOT NULL DEFAULT '0' COMMENT 'label:Mfg Suggested Retail Price;tag:MSRPrice;tab:$id;order:42'"); }
        // All of these fields have been moved to attributes or removed. Used for customization only.
//      if (!dbFieldExists(BIZUNO_DB_PREFIX."inventory", 'description_long')) { dbGetResult("ALTER TABLE ".BIZUNO_DB_PREFIX."inventory ADD description_long TEXT COMMENT 'type:textarea;label:Long Description;tag:DescriptionLong;tab:$id;order:10'"); }
//      if (!dbFieldExists(BIZUNO_DB_PREFIX."inventory", 'manufacturer'))     { dbGetResult("ALTER TABLE ".BIZUNO_DB_PREFIX."inventory ADD manufacturer VARCHAR(24) NOT NULL DEFAULT '' COMMENT 'label:Manufacturer;tag:Manufacturer;tab:$id;order:40'"); }
//      if (!dbFieldExists(BIZUNO_DB_PREFIX."inventory", 'model'))            { dbGetResult("ALTER TABLE ".BIZUNO_DB_PREFIX."inventory ADD model VARCHAR(24) NOT NULL DEFAULT '' COMMENT 'label:Model;tag:Model;tab:$id;order:41'"); }
//      if (!dbFieldExists(BIZUNO_DB_PREFIX."inventory", 'meta_keywords'))    { dbGetResult("ALTER TABLE ".BIZUNO_DB_PREFIX."inventory ADD meta_keywords VARCHAR(255) NOT NULL DEFAULT '' COMMENT 'label:Meta Keywords;tag:MetaKeywords;tab:$id;o rder:90;group:General'"); }
//      if (!dbFieldExists(BIZUNO_DB_PREFIX."inventory", 'meta_description')) { dbGetResult("ALTER TABLE ".BIZUNO_DB_PREFIX."inventory ADD meta_description VARCHAR(255) NOT NULL DEFAULT '' COMMENT 'label:Meta Description;tag:MetaDescription;tab:$id;order:91;group:General'"); }
        bizAutoLoad(BIZBOOKS_ROOT."controllers/inventory/admin.php", 'inventoryAdmin');
        $inv = new inventoryAdmin();
        $inv->installPhysicalFields();
    }
}
