<?php
/*
 * Functions to support API operations through Bizuno
 *
 * NOTICE OF LICENSE
 * This source file is subject to the Open Software License (OSL 3.0)
 * that is bundled with this package in the file LICENSE.TXT.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/OSL-3.0
 *
 * DISCLAIMER
 * Do not edit or add to this file if you wish to upgrade Bizuno to newer
 * versions in the future. If you wish to customize Bizuno for your
 * needs please refer to http://www.phreesoft.com for more information.
 *
 * @name       Bizuno ERP
 * @author     Dave Premo, PhreeSoft <support@phreesoft.com>
 * @copyright  2008-2022, PhreeSoft, Inc.
 * @license    http://opensource.org/licenses/OSL-3.0 Open Software License (OSL 3.0)
 * @version    6.x Last Update: 2021-12-02
 * @filesource /controllers/bizuno/import.php
 */

namespace bizuno;

class bizunoImport
{
    public $moduleID = 'bizuno';

    function __construct()
    {
        $this->lang = getLang('bizuno'); // needs to be hardcoded as this is extended by extensions
        $this->authReferers = [BIZUNO_SRVR, 'https://www.paypal.com/']; // list of authorized referers to access the api
    }

    /**
     * Main entry point structure for the import/export operations
     * @param array $layout - structure coming in
     * @return modified structure
     */
    public function impExpMain(&$layout=[])
    {
        if (!$security = validateSecurity('bizuno', 'impexp', 2)) { return; }
        $title= lang('bizuno_impexp');
        $data = ['title'=> $title,
            'divs'    => [
                'toolbar'=> ['order'=>20,'type'=>'toolbar','key' =>'tbImpExp'],
                'heading'=> ['order'=>30,'type'=>'html',   'html'=>"<h1>$title</h1>"],
                'biz_io' => ['order'=>60,'type'=>'tabs',   'key' =>'tabImpExp']],
            'tabs'=>[
                'tabImpExp'=>['divs'=>['module'=>['order'=>10,'type'=>'divs','label'=>lang('module'),'divs'=>['body'=>['order'=>50,'type'=>'tabs','key'=>'tabAPI']]]]],
                'tabAPI'   => ['styles'=>['height'=>'300px'],'attr'=>['tabPosition'=>'left', 'fit'=>true, 'headerWidth'=>250]]],
            'lang'    => $this->lang];
        $apis = getModuleCache('bizuno', 'api', false, false, []);
        msgDebug("\nLooking for APIs = ".print_r($apis, true));
        foreach ($apis as $settings) {
            $parts = explode('/', $settings['path']);
            if (empty(getModuleCache($parts[0], 'properties', 'path'))) { continue; }
            if (file_exists (getModuleCache($parts[0], 'properties', 'path')."/{$parts[1]}.php")) {
                $fqcn = "\\bizuno\\".$parts[0].ucfirst($parts[1]);
                bizAutoLoad(getModuleCache($parts[0], 'properties', 'path')."/{$parts[1]}.php", $fqcn);
                $tmp = new $fqcn();
                $tmp->{$parts[2]}($data); // looks like phreebooksAPI($data)
            }
        }
        portalMigrateGetMgr($layout);
        $layout = array_replace_recursive($layout, viewMain(), $data);
    }

    /**
     * Handles incoming requests through the API, for now just handles posting orders
     * @param array $layout - structure coming in, typically empty array
     * @return modified $layout
     */
    public function processRcv(&$layout=[])
    {
        $orderID = clean('rID', 'integer', 'get');
        if (!$this->validateReferer()) { return; }
        if (empty($orderID)) { // wordpress API to wordpress host is adding slashes to Order?? strip them, others?
            if (substr($_POST['Order'], 1, 1) == "\\") { $_POST['Order'] = stripslashes($_POST['Order']); }
            $order = clean('Order', 'json', 'post');
        } else {
            $order = $this->mapOrder($orderID);
        }
        if (empty($order)) { return msgAdd("Bad order ID!"); }
        $jID = getModuleCache('bizuno','settings','bizuno_api','auto_detect', 0);
        if (empty($jID)) { $jID = $this->getStockLevels($order['Item']); } // assumes customer order was placed
        $result = $this->apiJournalEntry($layout, $order, $jID);
        msgDebug("\nAfter post, Bizuno main record ID = $result");
        if (!empty($result)) { update_post_meta($orderID, 'bizuno_order_exported', 1); }
    }

    /**
     * Make sure the request is coming from within
     * @return type
     */
    private function validateReferer()
    {
        $access = false;
        foreach ($this->authReferers as $ref) {
            if (strpos($_SERVER['HTTP_REFERER'], $ref) === 0) { $access = true; }
        }
        return $access ? true : msgAdd("Illegal Access!", 'trap');
    }

    /**
     *
     * @param type $order_id
     * @param type $data
     * @return type
     */
    private function mapOrder($order_id, $data=[]) {
        $settings = $this->configGet(array('bizuno_api_user', 'bizuno_api_pw', 'bizuno_api_comp', 'bizuno_api_prefix', 'bizuno_api_prefix_cust'));
        $order_post = wpdbGetRow(PORTAL_DB_PREFIX.'posts', "ID='$order_id' AND post_type='shop_order'");
        msgDebug("\norder post = ".print_r($order_post, true));
        $order_post_meta = wpdbGetMulti(PORTAL_DB_PREFIX.'postmeta', "post_id='$order_id'", '', ['meta_key','meta_value']);
//        msgDebug("\norder post meta= ".print_r($order_post_meta, true));
        foreach ($order_post_meta as $meta_num ) { $order_info[$meta_num['meta_key']] = $meta_num['meta_value']; }
        $order_item = wpdbGetMulti(PORTAL_DB_PREFIX.'woocommerce_order_items', "`order_id`=$order_id AND `order_item_type` LIKE 'line_item'");
        msgDebug("\norder items= ".print_r($order_item, true));
        foreach ($order_item as $item_info) {
            $item_meta = wpdbGetMulti(PORTAL_DB_PREFIX.'woocommerce_order_itemmeta', "order_item_id = {$item_info['order_item_id']}", '', ['meta_key' , 'meta_value']);
            foreach ($item_meta as $row) { $item_formated[$row['meta_key']] = $row['meta_value']; }
            $order_item_meta[$item_info['order_item_name']] = $item_formated;
        } //this is a array ([order_item_id] => array ( [meta_key] => [meta_value] ))
        // Get the Shipping method, this may move depending on shipping plug-in used
        $shipping_method = wpdbGetValue(PORTAL_DB_PREFIX.'woocommerce_order_items', "order_item_name", "order_id=$order_id AND order_item_type='shipping'");
        $customer_id = $order_info['_billing_email'];
        $total_tax = $order_info['_order_tax']+$order_info['_order_shipping_tax']+$order_info['_cart_discount_tax'];
        msgDebug("\nMapping order = ".print_r($order_info, true));
        $Map = [
            'General' => [
//              'OrderID'         => $settings['bizuno_api_prefix'] . $order_id, // force a new invoice
                'PurchaseOrderID' => $settings['bizuno_api_prefix'] . $order_id,
                'OrderDate'       => substr($order_post['post_date'], 0, strpos($order_post['post_date'], ' ')),
                'OrderTotal'      => $order_info['_order_total'],
                'SalesTaxAmount'  => $total_tax,
//              'DiscountTotal'   => $this->fetchTotal('coupon', $totals_query->rows),
//              'SalesTaxPercent' => '',
//              'SalesTaxTitle'   => '',
                'ShippingTotal'   => $order_info['_order_shipping'],
                'ShippingCarrier' => $shipping_method, //$this->shipmentMap($order_info['_shipping_method']),
                'OrderNotes'      => $order_post['post_excerpt']],
            'Payment' => [
                'Method'          => $order_info['_payment_method'],
                'Title'           => !empty($order_info['_payment_method_title']) ? $order_info['_payment_method_title'] : $order_info['_payment_method'],
                'Status'          => $this->guessPmtStatus($order_info), // $order_info['_payment_cap_auth'],
                'Authorization'   => $order_info['_payment_auth_code'], // Authorization code from credit cards that need to be captured to complete the sale
                'TransactionID'   => $order_info['_payment_transaction_id'], // transaction ID from gateway
                'Hint'            => $order_info['_payment_card_hint']], // credit card hint, format 12**********7890
            'Billing' => [
                'CustomerID'      => $customer_id,
                'CompanyName'     => $order_info['_billing_company'] != '' ? $order_info['_billing_company'] : $order_info['_billing_first_name'].' '.$order_info ['_billing_last_name'],
                'Contact'         => $order_info['_billing_company'] == '' ? $order_info['_billing_first_name'].' '.$order_info['_billing_last_name'] : '',
                'Address1'        => $order_info['_billing_address_1'],
                'Address2'        => $order_info['_billing_address_2'],
                'City'            => $order_info['_billing_city'],
                'State'           => $order_info['_billing_state'],
                'PostalCode'      => $order_info['_billing_postcode'],
                'Country'         => clean($order_info['_billing_country'], ['format'=>'country', 'option'=>'iso3']),
                'Telephone'       => $order_info['_billing_phone'],
                'Email'           => $order_info['_billing_email']],
            'Shipping' => [
                'CompanyName'     => $order_info['_shipping_company'] != '' ? $order_info['_shipping_company'] : $order_info['_shipping_first_name'].' '.$order_info ['_shipping_last_name'],
                'Contact'         => $order_info['_shipping_company'] == '' ? $order_info['_shipping_first_name'].' '.$order_info['_shipping_last_name'] : '',
                'Address1'        => $order_info['_shipping_address_1'],
                'Address2'        => $order_info['_shipping_address_2'],
                'City'            => $order_info['_shipping_city'],
                'State'           => $order_info['_shipping_state'],
                'PostalCode'      => $order_info['_shipping_postcode'],
                'Country'         => clean($order_info['_shipping_country'], ['format'=>'country', 'option'=>'iso3']),
                'Telephone'       => $order_info['_billing_phone'],
                'Email'           => $order_info['_billing_email']],
            ];
        $Map['Item'] = [];
        if (!empty($order_item_meta)) { foreach ($order_item_meta as $item_name => $item_info) {
        //    $item_tax_total = $item['product_quantity'] * $item['tax'];
            $Map['Item'][] = [
                'ItemID'          => $this->fetchSKU($item_info['_product_id'], $item_info['order_item_name']),
                'Description'     => $item_name,
                'Quantity'        => $item_info['_qty'],
                'UnitPrice'       => $item_info['_line_subtotal']/$item_info['_qty'],
                'SalesTaxPercent' => '',
                'SalesTaxTitle'   => '',
                'SalesTaxAmount'  => $item_info['_line_tax'],
                'TotalPrice'      => $item_info['_line_total']];
        } }
        if (!empty($data)) { $Map['Other'] = $data; }
        return $Map;
    }

    public function migrateMgr(&$layout=[])
    {
        if (!$security = validateSecurity('bizuno', 'impexp', 4)) { return; }
        $modID = clean('modID','cmd', 'get');
        if (!$mod = portalMigrateLoad($modID)) { return msgAdd("migration module $modID not found!"); }
        $mod->manager($layout);
    }

    /**
     * API method to process incoming orders
     * @param array $layout - structure coming in
     * @param array $order - order data passed through the API
     * @param integer $jID [default 0] - used to preset the journal being written to
     * @return array modified $layout structure
     */
    public function apiJournalEntry(&$layout, $order=[], $jID=0)
    {
        msgDebug("\nWorking with submitted order = ".print_r($order, true));
        if (!dbConnected()) { return msgAdd('There was an issue connecting to your account! Please check your credentials.'); }
        $this->itemTotal= 0;
        $this->main     = ['journal_id'=>$jID, 'method_code'=>'','post_date'=>date('Y-m-d')];
        $this->items    = $map = [];
        $this->inStock  = true;
        require(BIZBOOKS_ROOT."controllers/bizuno/maps/journal.php"); // loads the journal map file structure
        $this->setJournalMain($map, $order);
        $this->setJournalItem($map, $order, $jID);
        $this->setJournalFreight($order, $jID);
        $this->setJournalTax($order, $jID);
        $this->setJournalDiscount($order, $jID); // must be last, also does unbalance correction for discrepancies between order total and item total
        $this->setJournalTotal($order, $jID);
        switch ($jID) {
            case 6:
                if (empty($this->main['invoice_num'])) { $this->main['waiting'] = 1; }
                break;
            case 10:
                $dup = dbGetValue(BIZUNO_DB_PREFIX.'journal_main', 'id', "journal_id=10 AND purch_order_id='{$this->main['purch_order_id']}'");
                if ($dup) { return msgAdd(sprintf(lang('err_gl_invoice_num_dup'), lang('reference'), $this->main['purch_order_id'])); }
                $this->main['waiting'] = 1; // force unshipped
                break;
            case 12:
                if (!empty($this->main['purch_order_id'])) {
                    $dup = dbGetValue(BIZUNO_DB_PREFIX.'journal_main', 'id', "journal_id=12 AND purch_order_id='{$this->main['purch_order_id']}'");
                    if ($dup) { return msgAdd(sprintf(lang('err_gl_invoice_num_dup'), lang('reference'), $this->main['purch_order_id'])); }
                }
                $this->main['waiting'] = 1; // force unshipped
                break;
        }
        if ( empty($this->main['gl_acct_id'])) { $this->main['gl_acct_id'] = $this->setDefGL($jID); }
        if (!empty($this->main['store_id']))   { // try to find the store record ID from the short_name
            $sID = dbGetValue(BIZUNO_DB_PREFIX.'contacts', 'id', "type='b' AND short_name='{$this->main['store_id']}'");
            if ($sID) { $this->main['store_id'] = $sID; }
        }
        $this->main['method_code'] = $this->guessShipMethod($this->main['method_code']); // try to map shipping method
        bizAutoLoad(BIZBOOKS_ROOT."controllers/phreebooks/journal.php", 'journal');
        // ***************************** START TRANSACTION *******************************
        dbTransactionStart();
        $journal = new journal(0, $jID, $this->main['post_date']);
        $journal->main = array_replace($journal->main, $this->main);
        // guess Sales Tax
        $journal->main['tax_rate_id'] = $this->taxGuess($order['General'], $order['General']['OrderTotal']); // try to determine tax rate at the order level
        if (!empty($journal->main['tax_rate_id'])) { $this->setTaxItems($journal->main['tax_rate_id']); } // changes $this->items
        $structure = dbLoadStructure(BIZUNO_DB_PREFIX.'journal_main', $jID);
        validateData($structure, $journal->main);
        $journal->items = $this->items;
        // Check to see if customer already exists in db
        $cID = dbGetContact($this->main, '_b');
        unset($this->main['short_name_b'], $journal->main['short_name_b']); // causes post errors as field is not in journal_main
        if ($cID) { // found match
            $journal->main['contact_id_b'] = $cID['contact_id'];
            $journal->main['address_id_b'] = $cID['address_id'];
        } else { // force the contact record to be created, set $_POST so contact info can be saved
            foreach ($journal->main as $key => $value) { $_POST[$key] = $value; } // @todo Temp until contacts module can be re-written to accept ->main and process
            $_POST['tax_rate_id_b'] = $journal->main['tax_rate_id'];
            $journal->updateContact_b = true;
            // set the contacts security to add contact if not currently at that level
            if (!$security = validateSecurity('contacts', 'mgr_c', 2)) { setUserCache('security', 'mgr_c', 2); }
        }
        msgDebug("\nReady to post order, main = ".print_r($journal->main, true));
        msgDebug("\nitems = ".print_r($journal->items, true));
        if (!$journal->Post()) { dbTransactionRollback(); return; }
        $this->setJournalPayment($map, $order, $journal->main['id']);
        // ***************************** END TRANSACTION *******************************
        dbTransactionCommit();
        $invoiceRef= pullTableLabel('journal_main', 'invoice_num', $journal->main['journal_id']);
        $billName  = isset($journal->main['primary_name_b']) ? $journal->main['primary_name_b'] : $journal->main['description'];
        msgAdd(sprintf(lang('msg_gl_post_success'), $invoiceRef, $journal->main['invoice_num']), 'success');
        msgLog('Bizuno API -'.lang('save')." $invoiceRef ".$journal->main['invoice_num']." - $billName (rID={$journal->main['id']}) ".lang('total').": ".viewFormat($journal->main['total_amount'], 'currency'));
        $_GET['rID'] = $journal->main['id']; // set the ID for hook processing
        return $journal->main['id'];
    }

    private function setDefGL($jID=0)
    {
        if (in_array($jID, [3,4,6,7])) { return getModuleCache('phreebooks', 'settings', 'vendors', 'gl_payables'); }
        return getModuleCache('bizuno','settings','bizuno_api','gl_receivables',getModuleCache('phreebooks','settings','customers','gl_receivables'));
    }

    private function setDefGLItem($jID=0)
    {
        if (in_array($jID, [3,4,6,7])) { return getModuleCache('phreebooks', 'settings', 'vendors', 'gl_purchases'); }
        return getModuleCache('bizuno', 'settings', 'bizuno_api', 'gl_sales', getModuleCache('phreebooks','settings','customers','gl_sales'));
    }

    /**
     * Maps the JSON input key/value pair to the Bizuno journal_main table format.
     * @param array $map - map of API keys to journal_main fields
     * @param array $order - data received through the API
     */
    private function setJournalMain($map, $order=[])
    {
        foreach ($map['General'] as $idx => $value) {
            if (isset($order['General'][$idx])) { $this->main[$value['field']]      = clean($order['General'][$idx], $value['format']); }
        }
        foreach ($map['Contact'] as $idx => $value) {
            if (isset($order['Billing'][$idx])) { $this->main[$value['field'].'_b'] = clean($order['Billing'][$idx], $value['format']); }
        }
        foreach ($map['Contact']as $idx => $value) {
            if (isset($order['Shipping'][$idx])){ $this->main[$value['field'].'_s'] = clean($order['Shipping'][$idx], $value['format']); }
        }
//      unset($this->main['short_name_b']); // commenting this out causes only an address check for cID, ignores test of the cID based on Contact ID
        unset($this->main['short_name_s']);
    }

    /**
     * Maps the JSON input key/value pair to the Bizuno journal_item table format.
     * @param array $map - map of API keys to journal_main fields
     * @param array $order - data received through the API
     */
    private function setJournalItem($map, $order=[], $jID=0)
    {
        $itmCnt = 1;
        foreach ($order['Item'] as $item) {
            $temp = ['item_cnt'=>$itmCnt,'gl_type'=>'itm','post_date'=>$this->main['post_date']];
            foreach ($map['Item'] as $idx => $value) {
                if (isset($item[$idx])) { $temp[$value['field']] = clean($item[$idx], $value['format']); }
            }
            if (empty($temp['gl_account'])) { $temp['gl_account'] = $this->setDefGLItem($this->main['journal_id']); }
            // the tax guess at item levels has been commented out as it is about impossible to guaranty accuracy and results in out of balance errors.
            $temp['tax_rate_id'] = $this->taxGuess($item, $temp['credit_amount']);
            $this->itemTotal+= $temp['credit_amount'];
            if (in_array($jID, [3,4,6,7])) {
                $temp['debit_amount'] = $temp['credit_amount'];
                unset($temp['credit_amount']);
            }
            $this->items[] = $temp;
            $itmCnt++;
        }
        msgDebug("\nFinished setJournalItem, inStock = ".($this->inStock?'TRUE':'FALSE'));
        // process any notes, this is after map so need to use table field names to inject no-sku item
        if (!empty($order['General']['OrderNotes'])) {
            $this->items[] = ['item_cnt'=>$itmCnt,'gl_type'=>'itm','qty'=>'1','sku'=>'','description'=>$order['General']['OrderNotes'], 'gl_account'=>$this->setDefGLItem($this->main['journal_id']),'post_date'=>$this->main['post_date']];
        }
    }

    /**
     * Sets the shipping item record
     * @param array $order - order as received through the API
     */
    private function setJournalFreight($order=[], $jID=0)
    {
        if (empty($order['General']['ShippingTotal'])) { return; }
        $this->items[] = [
            'qty'          => 1,
            'sku'          => '',
            'description'  => "title:".lang('shipping'),
            'gl_type'      => 'frt',
            'debit_amount' =>  in_array($jID, [3,4,6,7]) ? $order['General']['ShippingTotal'] : 0,
            'credit_amount'=> !in_array($jID, [3,4,6,7]) ? $order['General']['ShippingTotal'] : 0,
            'gl_account'   => getModuleCache('proLgstc','settings','general','gl_shipping_c',getModuleCache('phreebooks','settings','customers','gl_sales')),
            'post_date'    => $this->main['post_date']];
        $this->itemTotal += $order['General']['ShippingTotal'];
    }

    /**
     * Creates a tax item record making the assumption that the tax has been properly calculated at the cart
     * @param array $order - order as received through the API
     */
    private function setJournalTax($order=[], $jID=0)
    {
        if (empty($order['General']['SalesTaxAmount'])) { return; }
        $this->main['sales_tax']  = $order['General']['SalesTaxAmount'];
        $this->main['tax_rate_id']= getModuleCache('bizuno','settings','bizuno_api','tax_rate_id',0);
        $this->items[] = [
            'qty'          => 1,
            'sku'          => '',
            'description'  => "title:Sales Tax",
            'gl_type'      => empty($this->main['tax_rate_id']) ? 'glt' : 'tax', // if no rate found, add to order tax (Amount) so we don't have to worry about
            'debit_amount' => in_array($jID, [3,4, 6,13,20,21,22])       ? $order['General']['SalesTaxAmount'] : 0,
            'credit_amount'=> in_array($jID, [7,9,10,12,14,16,17,18,19]) ? $order['General']['SalesTaxAmount'] : 0,
            'gl_account'   => getModuleCache('bizuno','settings','bizuno_api','gl_tax'),
            'post_date'    => $this->main['post_date']];
        $this->itemTotal  += $order['General']['SalesTaxAmount'];
    }

    /**
     * Check item total to order total, any difference should be made into a discount record
     */
    private function setJournalDiscount($order=[], $jID=0)
    {
        $balanceCheck = $this->itemTotal - $this->main['total_amount'];
        if ($balanceCheck == 0) { return; }
        $this->items[] = [
            'qty'          => 1,
            'sku'          => '',
            'description'  => "title:".lang('discount'),
            'gl_type'      => 'dsc',
            'debit_amount' => !in_array($jID, [3,4,6,7]) ? $balanceCheck : 0,
            'credit_amount'=>  in_array($jID, [3,4,6,7]) ? $balanceCheck : 0,
            'gl_account'   => getModuleCache('bizuno','settings','bizuno_api','gl_discount'),
            'post_date'    => $this->main['post_date']];
    }

    /**
     * Creates the total item record
     * @param array $order - order as received through the API
     */
    private function setJournalTotal($order=[], $jID=0)
    {
        $this->items[] = [
            'qty'          => 1,
            'sku'          => '',
            'description'  => "title:".lang('total'),
            'gl_type'      => 'ttl',
            'debit_amount' => !in_array($jID, [3,4,6,7]) ? $order['General']['OrderTotal'] : 0,
            'credit_amount'=>  in_array($jID, [3,4,6,7]) ? $order['General']['OrderTotal'] : 0,
            'gl_account'   => $this->setDefGL($this->main['journal_id']),
            'post_date'    => $this->main['post_date']];
    }

    /**
     * Set the payment status of an order
     * @param array $map - map of API keys to journal table fields
     * @param array $order - order as received through the API
     * @param integer $rID - db record id of the journal entry if it exists
     * @return type
     */
    private function setJournalPayment($map, $order=[], $rID=0)
    {
        if ($order['General']['OrderTotal'] == 0) { return; } // for free orders, this is not necessary
        $pmtInfo = [];
        foreach ($map['Payment'] as $idx => $value) {
            if (isset($order['Payment'][$idx])) { $pmtInfo[$value['field']] = clean($order['Payment'][$idx], $value['format']); }
            else                                { $pmtInfo[$value['field']] = ''; }
        }
        msgDebug("\nWorking with payment array = ".print_r($pmtInfo, true));
        if (!$rID) { return; }
        $iID = dbGetValue(BIZUNO_DB_PREFIX."journal_item", ['id','description'], "ref_id=$rID AND gl_type='ttl'");
        $iID['description'] .= ";method:{$pmtInfo['method_code']};title:{$pmtInfo['title']};status:{$pmtInfo['status']};hint:{$pmtInfo['hint']}";
        switch ($pmtInfo['status']) {
            case 'auth':
                if (empty($pmtInfo['transaction_id'])) { $pmtInfo['transaction_id'] .= $pmtInfo['auth_code']; }
                if (empty($pmtInfo['transaction_id'])) {
                    msgAdd("The order has been authorized but the Authorization Code is not present, the payment for this order must be completed in Bizuno AND at the merchant website.", 'caution');
                }
                break;
            case 'cap':
                // This can be written but needs to know the payment method, fetch the order record
                // check to make sure it was posted successfully
                // make sure it was journal 12 NOT 10, if 10 flag as payment received but product not available???
                // build the save $this->main array, try to map the merchant to get gl_account and reference_id no need to cURL merchant
                // post it, close it as it is now paid
                msgAdd("The order has been paid at the cart, the payment for this order must be completed manually in Bizuno.", 'caution');
            case 'unpaid':
            default:
        }
        dbWrite(BIZUNO_DB_PREFIX."journal_item", ['description'=>$iID['description'],'trans_code'=>$pmtInfo['transaction_id']], 'update', "id={$iID['id']}");
    }

    /**
     * Sets the tax rate to every entry in the item table to a set value.
     * @param integer $taxID - db record id of the tax rate
     */
    private function setTaxItems($taxID=0)
    {
        foreach ($this->items as $idx => $row) {
            if ($row['gl_type'] <> 'itm') { continue; }
            if (empty($row['tax_rate_id'])) { $this->items[$idx]['tax_rate_id'] = $taxID; }
        }
    }

    /**
     * @param unknown_type $id
     * @param unknown_type $name
     * @return Ambigous <>|unknown
     */
    private function fetchSKU($id, $name) {
        $result = wpdbGetValue(PORTAL_DB_PREFIX.'postmeta', "meta_value","post_id=$id AND meta_key LIKE '_sku'", true);
        if ($result !== false) {
            return $result;
        }
        return $name;
    }

    /**
     * Determines if all products are in stock for auto journal determination
     * @param array $items - List of items in the order
     * @return integer - 12 (sales journal) if order can be filled, 10 (Sales Order journal) if not
     */
    private function getStockLevels($items=[])
    {
        $jID = 12;
        foreach ($items as $item) {
            if ($item['ItemID'] == '') { continue; }
            $inv = dbGetValue(BIZUNO_DB_PREFIX."inventory", ['inventory_type', 'qty_stock'], "sku='{$item['ItemID']}'");
            msgDebug("\nChecking stock levels, SKU = {$item['ItemID']} and qty = {$item['Quantity']} and stock level = {$inv['qty_stock']}");
            msgDebug("\nCOG_ITEM_TYPES = ".COG_ITEM_TYPES." and inventory_type = {$inv['inventory_type']}");
            if (strpos(COG_ITEM_TYPES, $inv['inventory_type']) !== false && $inv['qty_stock'] < $item['Quantity']) { $jID = 10; }
        }
        return $jID;
    }

    /**
     * Attempts to guess the payment status for the order, either cap (Sale) or auth (Authorize Only), defaults to cap
     * @param array $order_info - Order details
     * @return string - either cap or auth depending on the order payment status
     */
    private function guessPmtStatus($order_info) {
        switch ($order_info['_payment_method']) {
            default:
            case 'elavon': return isset($order_info['_payment_cap_auth']) ? $order_info['_payment_cap_auth'] : 'cap';
            case 'paypal': // works for WooCommerce included PayPal Standard payment method
                return isset($order_info['_paypal_status']) && $order_info['_paypal_status']=='completed' ? 'cap' : 'auth';
        }
    }

    /**
     * Tries to determine the tax rate based on the order data supplied
     * @param array $data - order data after being mapped to Bizuno API format
     * @param float $itemTotal = order total
     * @return integer - best guess, if possible, of the tax rate based on what is in the tax rate table
     */
    private function taxGuess($data=[], $itemTotal=0, $cID=0)
    {
        msgDebug("\nGuessing tax with itemTotal = $itemTotal and contactID = $cID");
        if (empty($this->taxRates)) {
            $this->taxRates = dbGetMulti(BIZUNO_DB_PREFIX.'tax_rates', "type='c'"); // load the tax rates, Not used as not reliable enough
        }
        // use indexes SalesTaxRate (percent), SalesTaxTitle (text), SalesTaxAmount (float) to try to match tax rate
        $taxShipping = getModuleCache('phreebooks', 'settings', 'general', 'shipping_taxed', 0);
        if ($taxShipping && $cID) { return dbGetValue(BIZUNO_DB_PREFIX.'contacts', 'tax_rate_id', "id=$cID"); }
        if       (isset($data['SalesTaxTitle'])  && $data['SalesTaxTitle']) {
            foreach ($this->taxRates as $row) {
                if ($row['title'] == $data['SalesTaxTitle']) {
                    msgDebug(" ... and returning with title found and id = {$row['id']}");
                    return $row['id'];
                }
            }
        } elseif (isset($data['SalesTaxPercent'])&& $data['SalesTaxPercent'] > 0) {
            foreach ($this->taxRates as $row) {
                if ($row['tax_rate'] == $data['SalesTaxPercent']) {
                    msgDebug(" ... and returning with percent found and id = {$row['id']}");
                    return $row['id'];
                }
            }
        } elseif (isset($data['SalesTaxAmount'])  && $data['SalesTaxAmount'] > 0) {
            $itemTotal -= !empty($data['SalesTaxAmount']) ? $data['SalesTaxAmount'] : 0; // subtract out the sales tax amount from the submitted total
            $percentTotal = round((($data['SalesTaxAmount'] / $itemTotal) * 100), getModuleCache('bizuno', 'settings', 'locale', 'number_precision', 2));
            msgDebug(" ... now working with calculated percent $percentTotal");
            foreach ($this->taxRates as $row) {
                $percentRate = round($row['tax_rate'], getModuleCache('bizuno', 'settings', 'locale', 'number_precision', 2));
                msgDebug(" ... id = {$row['id']} and percentRate = $percentRate");
                if (abs($percentTotal-$percentRate) < 0.02) {
                    msgDebug(" ... and returning with amount found and id = {$row['id']}");
                    return $row['id'];
                }
            }
        }
        msgDebug(" ... and returning with no tax id found!");
        return 0; // no tax data found
    }

    /**
     * Takes a stab at guessing the shipping method based on the title match (i.e. FedEx Ground)
     * @param string $method - shipping method as passed through the API
     * @return string - encoded shipping method, if found.
     */
    private function guessShipMethod($method='')
    {
        msgDebug("\nEntering guessShipMethod with method = $method");
        if (empty($method)) { return ''; }
        $carriers = getModuleCache('proLgstc', 'carriers');
        foreach ($carriers as $props) { // as $carrier => $props
//          msgDebug("\nCarrier $carrier and props = ".print_r($props, true));
            if (empty($props['settings']['services']) || !is_array($props['settings']['services'])) { return ''; }
            foreach ($props['settings']['services'] as $row) {
                if ($row['text']==$method) {
                    msgDebug("\nFOUND MATCH, EXITING with method_code: {$row['id']}");
                    return $row['id'];
                }
            }
        }
        return '';
    }

    /**
     * Pulls an WP option or array of options
     * @param type $const
     * @return type
     */
    private function configGet($const='') {
        if (is_array($const)) {
            $output = [];
            foreach ($const as $value) { $output[$value] = get_option($value); }
        } else {
            $output = get_option($const);
        }
        return $output;
    }

    /**
     * Removes common fields from the inventory db table
     *
     * NOTE: No fields are removed as they may be used by other modules
     */
    protected function removeStoreFields()
    {
        // DO not remove any fields as they may be used by another cart.
    }
}
