<?php
/*
 * WordPress Plugin - Functions related to logging in from portal
 *
 * NOTICE OF LICENSE
 * This source file is subject to the Open Software License (OSL 3.0)
 * that is bundled with this package in the file LICENSE.TXT.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/OSL-3.0
 *
 * DISCLAIMER
 * Do not edit or add to this file if you wish to upgrade Bizuno to newer
 * versions in the future. If you wish to customize Bizuno for your
 * needs please refer to http://www.phreesoft.com for more information.
 *
 * @name       Bizuno ERP
 * @author     Dave Premo, PhreeSoft <support@phreesoft.com>
 * @copyright  2008-2022, PhreeSoft, Inc.
 * @license    http://opensource.org/licenses/OSL-3.0 Open Software License (OSL 3.0)
 * @version    3.x Last Update: 2021-08-06
 * @filesource controllers/portal/functions.php
 */

namespace bizuno;

if (!defined('BIZUNO_KEY')) { define('BIZUNO_KEY', 'sbt8Ms2g8LfeJ67Z'); }

/**
 * Validates the credentials of a user to the database
 * @param string $email - username or email address.
 * @param string $pass - password
 * @param bool $verbose - display error message if true
 */
function biz_authenticate($email='', $pass='', $verbose=true) {
    $user = \wp_authenticate( $email, $pass );
    if ($verbose && is_wp_error($user)) { msgAdd($user->get_error_message()); }
    return !is_wp_error($user) ? true : false;
}

/**
 * Bizuno operates in local time. Returns WordPress safe date in PHP date() format if no timestamp is present, else PHP date() function
 * @param string $format - [default: 'Y-m-d'] From the PHP function date()
 * @param integer $timestamp - Unix timestamp, defaults to now
 * @return string
 */
function biz_date($format='Y-m-d', $timestamp=null) {
    if (!is_null($timestamp)) {
        $local = date( $format, $timestamp );
    } else {
        $local = \wp_date( $format );
    }
//    msgDebug("\nLocal date via type: $format with timestamp: $timestamp is $local, timestamp is: ".\wp_date( 'Y-m-d H:i:s' ));
    return $local;
}

/**
 * Validates the user is logged in and returns the creds if true
 */
function biz_validate_user() {
    global $mixer;
    $scramble = clean('bizunoSession', 'text', 'cookie');
    if (empty($scramble)) { // try to see if logged into WordPress
        $usr = \wp_get_current_user();
        return !empty($usr->user_email) ? [$usr->user_email] : false;
    }
    $creds = json_decode(base64_decode($mixer->decrypt(BIZUNO_KEY, $scramble), true));
    return !empty($creds[0]) ? $creds : false;
}

/**
 * Verifies the username and password combination for a specified user ID from the Bizuno tables mapped to the portal
 * @param mixed $email - Bizuno user table, if type = id, then integer, else email
 * @param string $pass - password to verify in the portal
 * @param string $verbose - whether to display error message if login fails
 * @return boolean - true on success, false otherwise
 */
function biz_validate_user_creds($email='', $pass='') {
    msgDebug("\nEntering biz_validate_user_creds with email = $email and pass = ".(!empty($pass) ? '****' : '--empty--'));
    $creds= ['user_login'=>$email, 'user_password'=>$pass, 'remember'=>false];
    $user = \wp_signon( $creds, false );
    if (is_wp_error($user)) { return msgAdd($user->get_error_message()); }
    wp_set_current_user($user->ID);
    wp_set_auth_cookie($user->ID);
    setUserCache('profile', 'email', $email);
    set_user_cookie($email);  // just register email since business ID is not known yet
    if (!empty(getUserCache('profile', 'biz_id'))) { return true; }
    setSession([$email, getUserCache('profile', 'biz_id')]);
    return true;
}

function bizSetUser($user, $bizID=1) { // $user->user_email
    $roleID= !empty( $_POST['bizbooks_role_'  .$bizID] ) ? $_POST['bizbooks_role_'.$bizID] : 0;
    update_user_meta( $user->ID, 'bizbooks_enable_'.$bizID, 1 );
    update_user_meta( $user->ID, 'bizbooks_role_'  .$bizID, $roleID );
    $biz   = portalGetBizIDVal($bizID);
    $dbTmp = new db($biz);
    if (!$dbTmp || !$dbTmp->connected) { return; }
    if (!userTableExists("{$biz['prefix']}users", $dbTmp)) { return; }
    $userID= $dbTmp->Execute("SELECT admin_id FROM {$biz['prefix']}users WHERE email='$user->user_email' LIMIT 1", 'row');
    if (empty($userID)) {
        $sql = "INSERT INTO {$biz['prefix']}users (`email`, `title`, `role_id`, `settings`) VALUES ('$user->user_email', '$user->display_name', $roleID,
            '".json_encode(['profile'=>['email'=>$user->user_email, 'biz_id'=>$bizID, 'role_id'=>$roleID]])."');";
        $dbTmp->Execute($sql, 'insert');
    }
}

function bizDelUser($user, $bizID=1) {
    delete_user_meta( $user->ID, 'bizbooks_enable_'.$bizID );
    delete_user_meta( $user->ID, 'bizbooks_role_'  .$bizID );
    // inactivate the users table record in the Bizuno users table
    $biz   = portalGetBizIDVal($bizID);
    $dbTmp = new db($biz);
    if (!$dbTmp || !$dbTmp->connected) { return; }
    if (!userTableExists("{$biz['prefix']}users", $dbTmp)) { return; }
    $userID= $dbTmp->Execute("SELECT admin_id FROM {$biz['prefix']}users WHERE email='$user->user_email' LIMIT 1", 'row');
    if (!empty($userID)) {
        $sql = "UPDATE {$biz['prefix']}users SET inactive='1' WHERE admin_id={$userID['admin_id']};";
//      $GLOBALS['bizDebug'][] = "bizDelUser executing sql = $sql";
        $dbTmp->Execute($sql, 'update');
    }
}

/**
 *
 * @global type $bizunoUser
 * @param array $creds
 */
function setSession($creds) {
    global $bizunoUser;
    msgDebug("\nEntering setSession");
    if (isset($creds[0])) { $bizunoUser['profile']['email']  = $creds[0]; }
    $bizunoUser['profile']['biz_id'] = isset($creds[1]) ? $creds[1] : 0;
    $GLOBALS['updateUserCache']  = 0;
    $GLOBALS['updateModuleCache']= [];
}

/**
 * Set the encrypted cookie for the current user
 * @global class $mixer - Encryption class
 * @param string $email - users email address
 * @param integer $bizID - business ID, always 1 for WordPress
 * @param integer $time - future time to set expiration in seconds
 */
function set_user_cookie($email='', $bizID=0, $time=false) {
    global $mixer;
    msgDebug("\nSetting user and session cookie with email = $email and bizID = $bizID and time = $time");
    bizSetCookie('bizunoUser', $email, time()+(60*60*24*7)); // 7 days
    if (empty($time)) { $time = 60*60; } // 1 hour
    $now = time();
    $cookie = $mixer->encrypt(BIZUNO_KEY, base64_encode("[\"$email\",$bizID,$now]"));
    bizSetCookie('bizunoSession', $cookie, $now+$time);
}

/**
 * Sets the default database to use for users that are not logged in to allow for customization
 */
function portalSetDB() {
}

function getWordPressEmail($userID='') {
    $user_info = get_userdata(get_current_user_id());
    if (empty($userID) && !empty($user_info->user_email)) { return $user_info->user_email; }
    return $userID;
}

function bizuno_get_locale() { return \get_locale(); }

/**
 * Fix for Google that changes the format of the language to en-US (underscore to dash)
 * @param string $iso
 */
function cleanLang(&$iso='en_US') {
    $gcln = str_replace('-', '_', $iso);
    $parts= explode('_', $gcln);
    $iso  = strtolower($parts[0]).'_'.strtoupper($parts[1]);
    if (strpos(getUserCache('profile', 'language'), '-') !== false) { setUserCache('profile', 'language', $iso); }
}

/**
 *
 * @param type $table
 * @param type $pdb
 * @return boolean
 */
function userTableExists($table, $pdb) {
    if (!$stmt = $pdb->Execute("SHOW TABLES LIKE '$table'")) { return false; }
    if (!$row  = $stmt->fetch(\PDO::FETCH_ASSOC)) { return false; }
    $value = array_shift($row);
    if (false === $value) { return false; }
    return ($value==$table) ? true : false;
}

function portalRead($table, $criteria='') { return dbGetRow($table, $criteria); }

function portalMulti($table, $filter='', $order='', $field='', $limit=0)  { return dbGetMulti($table, $filter, $order, $field, $limit); }

function portalExecute($sql)  { return dbGetResult ($sql); }

function portalWrite($table, $data=[], $action='insert', $parameters='') {
    if ('business'==$table) { return; }
    return dbWrite($table, $data, $action, $parameters);
}

function portalDelete($email, $bizID) {
    $user = get_user_by( 'email', $email );
    bizDelUser($user, $bizID);
}

function portalUpdateBiz() { }

/**
 * This method retrieves data from a remote server using cURL
 * @param string $url - url to request data
 * @param string $data - data string, will be attached for get and through setopt as post or an array
 * @param string $type - [default 'get'] Choices are 'get' or 'post'
 * @return result if successful, false (plus messageStack error) if fails
 */
function portalCurl($url, $data='', $type='get', $opts=[]) {
    $method = strtolower($type)=='get' ? 'GET' : 'POST';
    if ($method=='POST') {
        $args  = [
            'method'     => $method,
            'timeout'    => 90,
            'redirection'=> 5,
            'httpversion'=> '1.0',
            'blocking'   => true,
            'headers'    => !empty($opts['headers']) ? $opts['headers'] : [],
            'body'       => $data,
            'cookies'    => [],
            'sslverify'  => false,
        ];
        msgDebug("\nCalling WordPress cURL POST with url = $url and args = ".print_r($args, true));
        $response = wp_remote_post($url, $args);
    } else { // GET
        global $wp_version;
        $args = array(
            'timeout'    => 5,
            'redirection'=> 5,
            'httpversion'=> '1.0',
            'user-agent' => 'WordPress/'.$wp_version.'; ' . home_url(),
            'blocking'   => true,
            'headers'    => array(),
            'cookies'    => array(),
            'body'       => null,
            'compress'   => false,
            'decompress' => true,
            'sslverify'  => false,
            'stream'     => false,
            'filename'   => null
        );
        $req = !empty($data) ? $url.'?'.$data : $url;
        msgDebug("\nCalling WordPress cURL GET with url = $req and args = ".print_r($args, true));
        $response = wp_remote_get ($req, $args);
    }
    if (is_wp_error($response)) {
       $error_message = $response->get_error_message();
       msgDebug("\nError using WordPress cURL: $error_message");
       unset($response); // since it is an object
       $response['body'] = '';
    }
    msgDebug("\nReceived back from server: ".print_r($response['body'], true));
    return $response['body'];
}

/**
 * Sets the paths for the modules, core and extensions needed to build the registry
 * *** Sequence is important, do not change! ***
 * @param $coreOnly - [default false] set to true to only load core and custom modules
 * @return module keyed array with path the modules requested
 */
function portalModuleList($active=true) {
    $modCore = ['bizuno','contacts','inventory','payment','phreebooks','phreeform'];
    foreach ($modCore as $module) { $modList[$module] = BIZBOOKS_ROOT."controllers/$module/"; }
    portalModuleListScan($modList, BIZUNO_CUSTOM); // load custom modules
    $plugins = ['pro']; // ['pro', 'migrate', 'locale']; // extensions that have modules
    foreach ($plugins as $plugin) {
        $root = str_replace('bizuno-accounting', "bizuno-$plugin", BIZBOOKS_ROOT);
        if (is_dir("{$root}controllers/")) {
            portalModuleListScan($modList, "{$root}controllers/", $plugin, $active);
        } else { // legacy systems
            portalModuleListScan($modList, "{$root}modules/", $plugin, $active);
        }
    }
    msgDebug("\nReturning from portalModuleList with list: ".print_r($modList, true));
    return $modList;
}

function portalModuleListScan(&$modList, $path, $plugin=false, $active=true) {
    if (!is_dir($path)) { return; }
    if ( $plugin && $active && !portalValidatePlugin($plugin)) { return; }
    $custom = scandir($path);
    foreach ($custom as $name) {
        if ($name=='.' || $name=='..' || !is_dir($path.$name)) { continue; }
        if (file_exists($path."$name/admin.php")) { $modList[$name] = $path."$name/"; }
    }
}

/**
 * Fetches a list of valid businesses the user has access to
 * @return array
 */
function portalGetBizIDs() {
   $email  = getUserCache('profile', 'email');
   $userID = get_current_user_id();
   $output = [];
   foreach ($GLOBALS['bizPortal'] as $idx => $biz) {
       if (empty(get_user_meta( $userID, 'bizbooks_enable_'.$biz['id'], true ))) { continue; }
       $dbTmp = new db($biz);
       if (!$dbTmp || !$dbTmp->connected) {
           msgAdd("Failed connecting to the business: {$biz['title']}!");
           continue;
       }
       if (userTableExists("{$biz['prefix']}configuration", $dbTmp)) {
           $user   = $dbTmp->Execute("SELECT admin_id FROM {$biz['prefix']}users WHERE email='$email' AND inactive='0' LIMIT 1", 'row');
           if (empty($user)) { continue; } // no access to this business
           $values = $dbTmp->Execute("SELECT * FROM {$biz['prefix']}configuration WHERE config_key='bizuno' LIMIT 1", 'row');
           $setting= $values ? json_decode($values['config_value'], true) : [];
           $GLOBALS['bizPortal'][$idx]['title'] = $setting['settings']['company']['primary_name'];
           $src    = !empty($setting['settings']['company']['logo']) ? BIZBOOKS_URL_FS."&src={$biz['id']}/images/{$setting['settings']['company']['logo']}" : BIZUNO_LOGO;
           $action = "bizuno/portal/login&bizID={$biz['id']}";
       } else { // not installed, make sure they have admin privileges to install
           if( !current_user_can('administrator') ) { continue; }
           $src    = BIZUNO_LOGO;
           $action = "bizuno/portal/login&bizID={$biz['id']}";
       }
       $output["{$biz['id']}"] = ['id'=>$biz['id'], 'title'=>$biz['title'], 'src'=>$src, 'action'=>$action];
   }
   return $output;
}

/**
 *
 * @param type $bizID
 * @param type $idx
 * @return type
 */
function portalGetBizIDVal($bizID, $idx=false) {
    foreach ($GLOBALS['bizPortal'] as $dbData) {
        if ($bizID <> $dbData['id']) { continue; }
        return !empty($idx) ? $dbData[$idx] : $dbData;
    }
}

/**
 * Returns the pull down list of skins from the bizuno-skins plugin if installed and enabled.
 */
function portalSkins(&$skins=[]) {
    $path = BIZBOOKS_ROOT.'assets/skins/';
    $choices = scandir($path);
    foreach ($choices as $choice) {
        if (!in_array($choice, ['.','..']) && is_dir($path.$choice)) {
            $skins[] = ['id'=>$choice, 'text'=>ucwords(str_replace('-', ' ', $choice))];
        }
    }
}

/**
 * Returns the path for the requested skin
 * @param string $skin - Skin to locate
 * @return string - path the requested skin
 */
function portalSkinPath($skin='bizuno') {
    $path = BIZBOOKS_URL_ROOT.'view/easyUI/jquery-easyui/themes/';
    if (!in_array($skin, ['bizuno', 'default'])) { $path = BIZBOOKS_URL_ROOT.'assets/skins/'; }
    msgDebug("\nreturning from portalSkinPath with skin path = $path");
    return $path;
}

/**
 * Returns the pull down list of icons from the bizuno-icons plugin if installed and enabled.
 */
function portalIcons(&$icons=[]) {
    $path = BIZBOOKS_ROOT.'assets/icons/';
    $choices = scandir($path);
    foreach ($choices as $choice) {
        if (!in_array($choice, ['.','..']) && is_dir($path.$choice)) {
            $icons[] = ['id'=>$choice, 'text'=>ucwords(str_replace('-', ' ', $choice))];
        }
    }
}

/**
 * Returns the path for the requested icon set
 * @param string $icon - Icon set to locate
 * @return string - path the requested icon set
 */
function portalIconPath($icon='default') {
    $path = BIZBOOKS_ROOT.'view/icons/';
    if (!in_array($icon, ['default'])) { $path = BIZBOOKS_ROOT.'assets/icons/'; }
    msgDebug("\nReturning from portalIconPath with icon path = $path");
    return $path;
}

function portalIconURL($icon='default') {
    $path = BIZBOOKS_URL_ROOT.'view/icons/';
    if (!in_array($icon, ['default'])) { $path = BIZBOOKS_URL_ROOT.'assets/icons/'; }
    msgDebug("\nReturning from portalIconURL with icon path = $path");
    return $path;
}

function portalMigrateGetMgr(&$layout=[]) {
    $path = BIZBOOKS_ROOT."controllers/migrate";
    $mods = scandir($path);
    foreach ($mods as $fn) {
        if ($fn=='.' || $fn=='..' || !is_dir($path.$fn)) { continue; }
        require_once($path."$fn/admin.php");
        $fqdn  = "\\bizuno\\{$fn}Admin";
        $ctrl  = new $fqdn();
        $ctrl->impExpMain($layout);
    }
}

function portalMigrateLoad($modID='') {
   $path = BIZBOOKS_ROOT."controllers/migrate/$modID/admin.php";
   if (!bizAutoLoad($path, "{$modID}Admin")) { return msgAdd("class $modID not found as path $path"); }
   $fqdn  = "\\bizuno\\{$modID}Admin";
   return new $fqdn();
}

/**
 * Validates that the user has permission to access the Plugin and it is activated
 * @return boolean
 */
function portalValidatePlugin($name) {
    if (get_option("bizuno_{$name}_active")) { return true; }
    require_once(ABSPATH.'wp-admin/includes/plugin.php'); // load is_plugin_active function
    if (\is_plugin_active("bizuno-$name/bizuno-$name.php")) { return true; }
    return false;
}
